(     Title:  Garbage collected record stack
       File:  gcrstack.fs;
     Author:  David N. Williams
    Version:  0.8.4
    License:  LGPL
  Test file:  gcrstack-test.fs
   Log file:  gcrstack.log
    Started:  January 19, 2011
    Revised:  April 23, 2011
)
\ Copyright (C) 2011 David N. Williams
(
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or at your option any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA 02111-1307 USA.

If you take advantage of the option in the LGPL to put a
particular version of this library under the GPL, the author[s]
would regard it as polite if you would put any direct
modifications under the LGPL as well, and include a copy of this
request near the beginning of the modified library source.  A
"direct modification" is one that enhances or extends the
library in line with its original concept, as opposed to
developing a distinct application or library which might use it.

This library implements a stack of addresses of records of fixed
size, together with a garbage collected record space.  It is a
variant of dynamic string and big number stacks developed
earlier by the author.

The design has in mind the three record types in the GNU
Multiple Precision library, gmp, for integer, rational, and
floating-point numbers, used to manage their memory allocation.
That library has functions for each type that initialize the
records and release their associated memory, which are
accomodated here in the >GINIT-REC-XT and >GFREE-REC-XT fields
in the the record space header.  Initialization in gmp not only
allocates initial memory, but sets the number to zero.

A copy operation is accommodated by the header field
>GCOPY-REC-XT.
)
decimal

[UNDEFINED] --- [IF]
: ---  ( -- )  \ nonnesting, nonwhitespace delimited terminator
  BEGIN
    source >in @ /string s" ---" search 0=
  WHILE
    2drop refill 0= ABORT" ***missing --- terminator"
  REPEAT
  2 + source nip + >in ! drop ; immediate
[THEN] 


\ *** TABLE OF CONTENTS
---
1 TERMINOLOGY

2 UTILITY WORDS

3 HEADER STRUCTURE
  >/gbuf  >/grec  >gfree-rec-xt  >gcopy-rec-xt
  >gbrkp  >gsp  >gsp0  >ggbg  >ggc-lock
  >gextra1 >gbuf /gspace-header

4 RECORD SPACE
  0gspace  make-gspace  free-grecs
  is-grec  in-gs
  collect-ggarbage ggarbage?
  ggc-off  ggc-on  ggc-lock@  ggc-lock!
  gunused  groom?  ?GROOM  ?CELL-GROOM
  /gstack  ?GSTACK-UNDER  ?GSTACK-CELL-UNDER  ?GSTACK-2CELLS-UNDER

5 STACK SUPPORT
  gsp--  gsp++  gdepth  (gpush)

6 STACK TRANSFER
  gs@  gs2@  gs3@  (>gs)  ?GEXTERN  >gs  (gs>)

7 LOAD AND STORE
  g,ra gs>extern GVARIABLE g@  g!

8 STACK MANIPULATION
  gdrop  gdup  gswap  gnip  gover  gtuck
  g2drop  g3drop  g2dup  gpick  gexchange  grot  -grot
---

\ *** 1 TERMINOLOGY
---
rec:  The basic data instance, a record consisting of a fixed
  number of cells for each data type, which contain information
  about the type and its possible memory allocation.

ra:  Short for "record address".
  
gstack:  Short for "garbage collected record stack", a LIFO
  stack, located in the record space, which also hold ra's
  subject to garbage collection.

gvariable:  Short for "garbage collected record variable".  A
  gvariable holds an ra, and is initialized when created to
  zero, the null ra.

gspace:  A space of records of a given type, including a header
  and the record stack.
---

\ *** 2 UTILITY WORDS

true [IF] \ DEBUGGING
[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ; [THEN]
: .#  ( n -- )  cr ." #" . cr ;
[UNDEFINED] [FALSE] [IF]
: [FALSE]  ( -- false )  false ; immediate [THEN]
[UNDEFINED] [TRUE] [IF]
: [TRUE]  ( -- true )   true  ; immediate [THEN]
[THEN]

[UNDEFINED] cell [IF]
1 cells constant cell [THEN]

[UNDEFINED] -cell [IF]
-1 cells constant -cell [THEN]

[UNDEFINED] cell/ [IF]
: cell/  ( cell*[x] -- x )  [ 1 cells ] literal / ; [THEN]

[UNDEFINED] cell- [IF]
: cell-  ( addr -- addr-cell )  [ 1 cells ] literal - ; [THEN]

[UNDEFINED] OFF [IF] : OFF  ( addr -- )  false swap ! ; [THEN]
[UNDEFINED] ON  [IF] : ON   ( addr -- )  true swap ! ;  [THEN]

[UNDEFINED] rdrop [IF]
: rdrop  ( r: a -- )  postpone r> postpone drop ; immediate [THEN]

[UNDEFINED] 3@ [IF]
: 3@  ( addr -- x3 x2 x1 ) dup >r [ 2 cells ] literal + @ r> 2@ ; [THEN]

: /STRUCT:  ( <"sizename"> u -- )  aligned constant ;

: cell:  ( offset <"name">  -- a-offset+1cell )       \ 200x FIELD:
  create aligned dup , cell+
DOES>  ( struc -- struc+offset )  @ + ;

: after:  ( offset <"name">  -- a-offset )
  create aligned dup ,
DOES>  ( struc -- struc+offset ) @ + ;

[UNDEFINED] @EXECUTE
[IF] [DEFINED] PERFORM
  [IF] [DEFINED] synonym
    [IF] synonym @EXECUTE PERFORM
    [ELSE] [DEFINED] alias
      [IF] ' PERFORM alias @EXECUTE
      [ELSE] cr .( ***can't use @EXECUTE) cr ABORT
      [THEN]
    [THEN]
  [ELSE] : @EXECUTE  @ EXECUTE ;
  [THEN]
[THEN]

\ *** 3 HEADER STRUCTURE
---
The minor structure is the grecord, consisting of a back link
field of one cell, followed by the actual record, of fixed size
depending on the type of gstack.  The structure of the actual
record is opaque.

The major structure is the gspace header.  The addresses of
current gspaces are held in VALUE's, one for each type of
gstack.  The type is parametrized by the record size and
the xt's for record initialization, release, and copying.
---
0  cell: >/gbuf        \ size of gspace buffer including stack
   cell: >/grec        \ record size including blink field
   cell: >gfree-rec-xt \ xt for releasing record
   cell: >gcopy-rec-xt \ xt for initializing and copying to target record 
   cell: >gbrkp        \ next available for record node
   cell: >gsp          \ gstack pointer
   cell: >gsp0         \ initial gstack pointer
   cell: >ggbg         \ true means there is garbage
   cell: >ggc-lock     \ true means collection not permitted
   cell: >gextra1      \ user slot #1
  after: >gbuf         \ buffer starts here
/STRUCT: /gspace-header


\ *** 4 RECORD SPACE
---
Each gspace has two parts, a header, followed by a buffer.  The
buffer contains a sequence of grecords, starting just after the
header and building up to just before the break.  It also
contains the gstack of record addresses (not grecord addresses),
building down to the break.
---

: 0gspace  ( 'gspace -- )
  ---
    Clear the gpace buffer and gstack in the gspace starting at
    'gspace.  Any external variables pointing to addresses in
    the gspace buffer are left pointing to limbo.  This may be
    executed with the gspace in an invalid state, as long as the
    /GBUF field of its gspace header is intact.

    NOTE: This word sets pointers.  It does not zero fill the
    gspace buffer, nor does it execute the xt for freeing
    records.  It ignores the >GEXTRA1 field in the header.
  ---
  ( 'gspace) >r
  r@ >gbuf r@ >gbrkp !
  r@ >gbuf r@ >/gbuf @ +
  ( offend) dup r@ >gsp !
            dup r@ >gsp0 !
            off  \ gstack underflow guard  
  false r@ >ggbg !   false r> >ggc-lock !
;

: make-gspace  ( free.xt copy.xt /rec /buf -- 'gspace )
  ---
    Allocate and initialize a gspace with /buf address units
    available for the gspace buffer, including the gstack.  The
    size is rounded up to address alignment, and the buffer
    begins and ends with cell alignment.  Return 'gspace, the
    address of the gspace.  The standard word FREE with 'gspace
    as input can be used to release the space, once any
    nongarbage grecord has been released.
  ---
  ( /buf) aligned dup
  /gspace-header + cell+ allocate throw  \ extra guard cell 
  ( 'gspace) >r
  r@ >/gbuf !          cell+ r@ >/grec !
  r@ >gcopy-rec-xt !   r@ >gfree-rec-xt !   r@ >gextra1 off
  r@ 0gspace r> ;

: gcopy-rec  ( src.ra targ.ra 'gspace -- )  \ includes target initialization
  postpone >gcopy-rec-xt postpone @EXECUTE ; immediate

: gfree-rec  ( ra 'gspace -- )
  postpone >gfree-rec-xt postpone @EXECUTE ; immediate

: free-grecs  ( 'gspace -- )
  ---
    Execute the xt in the GREC-FREE-XT field of the gspace
    header on all nongarbage grecords.  Set all external
    variables pointing to such records to zero.  Finally,
    execute 0GSPACE.

    NOTE: This word must be executed only when the gspace is in
    a valid state.
  ---
  >r r@ >gfree-rec-xt @   r@ >/grec @
     r@ >gbrkp @          r@ >gbuf
  r>
  LOCALS| 'gspace next break /grec free-xt |

  BEGIN
    next break u<
  WHILE
    next @ ( backlink)
    IF next cell+ free-xt EXECUTE next @ off THEN
    next /grec + to next
  REPEAT
  'gspace 0gspace
;

: collect-ggarbage  ( 'gspace -- collected? )
  ---
    The garbage flag in the gspace header is tested, and if
    there is garbage, it is collected, unless garbage collection
    is locked.  Garbage records are marked by null backward
    links.  Nongarbage records are bound by their backward
    links, pointing either to a variable, to an entry on the
    gstack [the deepest if there are several identical
    references].

    Garbage collection fills the gaps occupied by garbage
    records by moving any nongarbage records to lower memory one
    at a time.  The backward link of a record that is moved does
    not change, but the forward links, in at most one variable
    and/or possibly several gstack entries, are updated to point
    to the new record address.  This algorithm is "fast" because
    the backward links make it unnecessary to scan a list of
    variables, and because no record is moved more than once. 
    It does, however, require a scan of the gstack for each
    record that moves.

    When there is no garbage to collect, this word returns
    false.  If there is garbage, it throws an error when garbage
    collection is locked.  Otherwise it collects the garbage and
    returns true.

    We do not think it would be a good idea to factor this long
    word.
  ---
  dup >ggbg @ 0= IF drop false EXIT THEN
  dup >ggc-lock @ ABORT" ***record garbage locked"
  >r false r@ >ggbg !
  r@ >gsp0 @
  r@ >gsp @     r@ >gbrkp    dup @
  r@ >/grec @   r@ >gbuf
  0  r>
  LOCALS| header target next /grec break breakp sp sp0 |

\ locate first garbage hole (no need to check off end, because
\ we know there is a hole)

  BEGIN
    next @ ( backlink)
  WHILE  \ not garbage, skip to next
    next /grec + to next
  REPEAT
  next to target

\ garbage hole found

  BEGIN
    next break u<
  WHILE

  \ skip over garbage

    BEGIN
      next @ ( backlink) 0=   next break u< and
    WHILE
      next /grec + to next
    REPEAT

  \ move and update until next garbage

    BEGIN
      next @ ( backlink)   next break u< and

    WHILE  \ not garbage and not off end

    \ We always update the forward link pointed to by the
    \ backward link.  The backward link points to one of a
    \ variable or a deepest gstack entry.

      target cell+ next @ ( backlink)  !

    \ Scan the gstack and update copies of the old forward link.
    \ Note that if the backward link points into the gstack,
    \ we've already updated the deepest reference, which won't
    \ be found in the scan.

      sp0 sp ?DO
	i @ next cell+ =
	IF target cell+ i ! THEN
      cell +LOOP

    \ move grecord to hole

      next target /grec move
      next /grec + to next   target /grec + to target

    REPEAT  \ move and update
  REPEAT    \ garbage hole found

  target breakp !   true
;

: ggarbage?  ( 'gspace -- flag )
  ---
    Leave true if there is garbage in the gspace.  Not
    normally used by applications, since garbage collection is
    transparent.
  ---
  >ggbg @ ;

: make-ggarbage  ( ra 'gspace -- )
  ---
    Assume that ra is internal and not referenced by any forward
    link.  Make it into garbage.
  ---
  >r dup cell- off ( ra) r@ gfree-rec r> >ggbg on ;
0 [IF] \ debugging version
   LOCALS| 'space ra |
   ra cell- off
   ra 'space gfree-rec
   ra 'space >/grec @ cell- 0 fill
   'space >ggbg on ;
[THEN]

: ggc-off  ( 'gspace -- )
  ---
    Disable garbage collection in the gspace.  An error will be
    thrown if garbage collection is attempted.
  ---
  >ggc-lock on ;

: ggc-on  ( 'gspace -- )
  ---
    Enable garbage collection in the gspace.  That is the
    default.
  ---
  >ggc-lock off ;

: ggc-lock@  ( 'gspace -- flag )
  ---
    Fetch the gspace garbage collection "off" state.  Intended
    for saving the off state for later restoration after a usage
    of GGC-ON or GGC-OFF.
  ---
  >ggc-lock @ ;

: ggc-lock!  ( flag 'gspace -- )
  ---
    Set the gspace garbage collection "off" state according to
    flag.  Intended for restoring the off state previously
    fetched by GGC-LOCK@.
  ---
  >ggc-lock ! ;

: gunused  ( 'gspace -- u )
  ---
    Leave the number of address units available for grecords
    and gstack entries in the gspace buffer.
  ---
  dup >gsp @ swap >gbrkp @ - ;

: groom?  ( size 'gspace -- flag )
  ---
    Leave true if there is enough aligned room in the gspace for
    size address units starting at the break.
  ---
  >r r@ >gbrkp @ + aligned r> >gsp @ u<= ;

: ?GROOM  ( size 'gspace -- )
  ---
    Abort if there is not enough aligned room in the gspace for
    size address units starting at the break, even after garbage
    collection.
  ---
  2dup groom? IF 2drop EXIT THEN
  dup collect-ggarbage ( collected?) drop
  groom? 0= ABORT" ***record space overflow"
;

: ?CELL-GROOM  ( 'gspace -- 'gspace )  cell over ?GROOM ;

: is-grec  ( addr 'gspace -- flag )
  ---
    Leave true if the address points into the gspace buffer
    before the break, else false.
  ---
  dup >gbuf swap >gbrkp @ within ;

: in-gs  ( addr 'gspace -- flag )
  ---
    Leave true if the address points into the gstack, else
    false.
  ---
  dup >gsp @ swap >gsp0 @ within ;

: /gstack  ( 'gspace -- #au )  dup >gsp0 @ swap >gsp @ - ;

: ?GSTACK-UNDER  ( #au 'gspace -- )
  /gstack u> ABORT" ***record stack underflow" ;

: ?GSTACK-CELL-UNDER  ( 'gspace -- 'gspace )  cell over ?GSTACK-UNDER ;

: ?GSTACK-2CELLS-UNDER  ( 'gspace -- 'gspace )
  [ 2 cells ] literal over ?GSTACK-UNDER ;


\ *** 5 STACK SUPPORT

: gsp--  ( 'gspace -- )  >gsp -cell swap +! ;
: gsp++  ( 'gspace -- )  >gsp  cell swap +! ;

: gdepth  ( 'gspace -- u )
  ---
    Leave the number of items on the gstack.
  ---
  /gstack cell/ ;

: (gpush)  ( ra 'gspace -- ) ( g: -- ra )
  ---
    Pop a record address from the data stack and push it onto
    the gstack, with no check for stack room.  A check that
    could trigger garbage collection must be done before using
    this word.  If ra is not external, it must be supplied after
    the check, from the gstack or a variable, so it remains
    valid after a garbage collection.
  ---
  dup gsp-- >gsp @ ! ;


\ *** 6 STACK TRANSFER
---
WARNING:  The words GS@, GS2@, and GS3@ move items from the
gstack to the data stack.  If any of the records is bound, it
may move at the next garbage collection, making the data stack
ra invalid.  One way to avoid this is to invoke ?GROOM with a
sufficient size beforehand.  Another is to sandwich sections of
code where this could occur between GGC-OFF and GGC-ON.
---

: gs@  ( 'gspace -- ra ) ( g: ra -- ra )
  ---
    Leave the gstack unchanged, and copy the top ra to the data
    stack.  Throw an error if there is nothing on the gstack.
  ---
  ?GSTACK-CELL-UNDER >gsp @ @ ;

: gs2@  ( 'gspace -- ra1 ra2 ) ( g: ra1 ra2 -- ra1 ra2 )
  ---
    Leave the gstack unchanged, and copy the top two ra's to the
    data stack.  Throw an error if there are not two items on
    the gstack.  Useful for efficient processing of two record
    arguments.
  ---
  ?GSTACK-2CELLS-UNDER >gsp @ 2@ ;

: gs3@  ( 'gspace -- ra1 ra2 ra3 ) ( g: ra1 ra2 ra3 -- ra1 ra2 ra3 )
  ---
    Leave the gstack unchanged, and copy the top three ra's to
    the data stack.  Throw an error if there are not three items
    on the gstack.  Useful for efficient processing of three
    record arguments.
  ---
  [ 3 cells ] literal over ?GSTACK-UNDER >gsp @ 3@ ;

: (>gs)  ( ra 'gspace -- g: ra )
  ---
    Pop the external ra from the data stack and push it onto the
    gstack after checking for room, invoking garbage collection
    if necessary.  NOT TO BE USED with a bound record because a
    garbage collection could invalidate its address.
  ---
  ?CELL-GROOM (gpush) ;

: ?GEXTERN  ( ra 'gspace -- )
  ---
    The word >GS uses this on the ra because the push, or push
    with copy, operation could generate a garbage collection,
    which could invalidate ra.  That requirement can lifted, but
    that's complicated, and it's normally bad practise to push
    nonexternal records from the data stack.
  ---
  is-grec ABORT" ***record not external" ;

: >gs  ( ra 'gspace -- g: ra )
  ---
    Push the external record onto the gstack, without copying
    the record data into the gbuffer.  An error is thrown if the
    record is bound, or if there is not enough room on the
    gstack, even after garbage collection.
  ---
  2dup ?GEXTERN (>gs) ;

: new-grec  ( 'gspace -- ra )
  ---
    Create an uninitialized new grec with zero back link,
    without checking for room.  Care is needed with this word to
    ensure there is room and that no garbage collection occurs
    before it gets bound.

    Note:  Although the new grec is technically garbage, the
    garbage flag is not changed.
  ---
  dup >gbrkp dup @ cell+
  LOCALS| ra brkp 'gspace |
  brkp @ OFF
  'gspace >/grec @ brkp +!
  ra
;

: >gs-copy  ( ra 'gspace -- ) ( g: -- ra' )
  ---
    Copy the record data at the external ra on the data stack
    into the gbuffer, push it onto the gstack, and bind it to
    the top.  Errors are thrown if the source record is not
    external or if there is not enough room in gspace, even
    after garbage collection.  This word executes GCOPY-REC,
    which includes initialization.
  ---
  2dup ?GEXTERN
  dup >r >/grec @ cell+ r@ ?GROOM  \ now gc safe
  r@ new-grec                      \ not initialized
  ( ra') dup r@ (gpush)
  r@ >gsp @ over cell- !           \ back link
  ( ra ra') r> gcopy-rec           \ initialized and copied
;

: (gs>)  ( 'gspace -- ra ) ( g: ra -- )
  ---
    Throw an error if there is nothing on the gstack.  Drop the
    ra from the gstack without checking whether it's a grec
    which should be garbaged, and leave it on the data stack.
  ---
  ?GSTACK-CELL-UNDER dup >gsp @ @ swap gsp++
;

: (gdrop)  ( 'gspace -- ) ( g: ra -- )
  ---
    Drop ra from the gstack without checking for underflow.
  ---
  dup >gsp @ @
  LOCALS| ra 'gspace |
  ra 'gspace is-grec
  IF 'gspace >gsp @ ra cell- @ =
  IF ra 'gspace make-ggarbage THEN THEN
  'gspace gsp++
;


\ *** 7 LOAD AND STORE
---
There is no word GCONSTANT.  Instead, GS>EXTERN is used as a
helper for constructing defining words that make constants for
gspace instances.  See TCONSTANT in gcrstack-test.fs for an
example.
---

: GVARIABLE  ( "name" -- )
\ name does  ( -- dfa )
  ---
    GVARIABLE's hold ra's of any type, and are initialized to
    zero for the null pointer.  The assumption that no legal
    address can be zero is an implementation dependency.
   ---
  create 0 , ;

: g,ra  ( 'space -- ra ) ( g: b -- )
  ---
    Copy the b record data into a new aligned data space record,
    drop b from the gstack and leave the new ra.
  ---
  align here
  LOCALS| ra 'space |
  'space >/grec @ cell- allot
  'space gs@ ra 'space gcopy-rec 
  'space ?GSTACK-CELL-UNDER (gdrop) ra
;

: gs>extern  ( 'space -- ra.ext ) ( g: b -- )
  ---
    If the b record data is external, leave its ra, else leave
    the ra of a new, aligned copy of the data in data space.
  ---
  dup gs@ over is-grec ( 'space flag)
  IF g,ra ELSE (gs>) THEN ( ra.ext)
;

: g@  ( gvar.dfa 'gspace -- ) ( g: -- ra )
  ---
    Leave the ra held by the gvariable.  The ra may be either
    bound or external.
  ---
   ?CELL-GROOM swap @ swap (gpush) ; 

: g!  ( gvar.dfa 'gspace -- ) ( g: ra -- )
  ---
    Store the ra in the gvariable whose dfa is on the data
    stack.

    The only situation in which G! copies the record itself is
    when it is bound to another gvariable.  In that case, a new
    copy is generated and stored in the variable.  In
    particular, external records are not copied.

    Here's the logic.

    1. If the ra already in the variable [old.ra] is the same
       as the ra on the gstack [new.ra], increment the gstack
       pointer and exit.

    2. Consider old.ra first, because its record might get
       marked as garbage, leaving potentially more room in case
       the new record gets copied into gspace.

       if old.ra is bound
         if it's on the gstack, set its back link to the deepest
           entry
         else make it garbage
         endif
       endif

    3. Then consider new.ra.

       if new.ra is bound
         if it's bound to the gstack, set its back link
           to gvar.dfa
         else it's bound to another variable, collect garbage if
           room is needed, copy it with initialization into a
           new grec, set the copy's back link to gvar.dfa, and
           replace new.ra by the copy's ra
         endif
       endif

    4. Store new.ra at gvar.dfa and increment the gstack
       pointer.
  ---
  ?GSTACK-CELL-UNDER
  ( var 'gspace) 0 0 0 0

  LOCALS| blfa /grec new.ra old.ra 'gspace var |

  var @ to old.ra   'gspace >gsp @ @ to new.ra
  new.ra old.ra = IF 'gspace gsp++ EXIT THEN
  'gspace >/grec @ to /grec

  old.ra 'gspace is-grec IF  \ old record is bound
    0 ( next)
    'gspace >gsp0 @ 'gspace >gsp @ cell+
    ?DO
      i @ old.ra = IF ( prev.next) drop i ( next) THEN
    cell +LOOP
    ( next) dup IF
      old.ra cell- ( blfa) !
    ELSE drop old.ra 'gspace make-ggarbage THEN
  THEN

  new.ra 'gspace is-grec IF  \ new record is bound

    new.ra cell- ( new.blfa) dup @ 'gspace in-gs
 
    IF    \ it's bound to the gstack

      ( new.blfa) to blfa

    ELSE  \ it's bound to a different variable, copy it

      ( new.blfa) drop
      /grec 'gspace ?GROOM

      'gspace >gbrkp @ to blfa    \ in case there was gc

      'gspace >gsp @ @ ( src.ra)  \ in case there was gc
          blfa cell+
          'gspace gcopy-rec
      blfa cell+ to new.ra        \ for the variable
      blfa /grec + 'gspace >gbrkp !

    THEN
 
    var blfa !

  THEN

  new.ra var !   'gspace gsp++
;


\ *** 8 STACK MANIPULATION
---
From now on, the notation in gstack comments often does not
explicitly mention "ra".  We think of the stack item either as
the representation of the data type in the record contents, or
more literally as an ra, as appropriate.
---

: gdrop	 ( g: a -- )
  ---
    Abort if the gstack would underflow when popped.

    Otherwise increment the gstack pointer, thus popping the
    gstack.

    If the input ra is a grecord bound to the top of the gstack,
    mark it as garbage by setting its back link to NULL, set the
    garbage flag, and perform GFREE-REC-XT on it.
  ---
  ?GSTACK-CELL-UNDER (gdrop) ;

: gdup  ( 'space -- ) ( g: a -- a a )
  ---
    Leave a copy of the topmost gstack entry.  The record value
    is not copied.
  ---
  ?GSTACK-CELL-UNDER dup >gsp @ swap g@ ;

: gswap  ( 'space -- ) ( g: a b -- b a )
  ---
    Exchange the two most accessible entries on the gstack.
    Throw an error if there are less than two entries.  Neither
    record value is copied.
  ---
  >r [ 2 cells ] literal r@ ?GSTACK-UNDER
  r@ >gsp @ ( 'b) dup @ ( b)
  over cell+ ( 'a) dup @ ( a) r>

  LOCALS| 'space a 'a b 'b |

\ don't change blinks to gvars, nor to deeper than 'deep

  a b = IF EXIT THEN
  b 'a !   a 'b !

  a 'space is-grec IF
    a cell- @ ( blink) 'a =
    IF 'b a cell- ( blfa) ! THEN
  THEN

  b 'space is-grec IF
    b cell- @ ( blink) 'b =
    IF 'a b cell- ( blfa) ! THEN
  THEN
;

: gnip  ( 'gspace -- ) ( g: a b -- b )
  ---
    Drop the next to top item from the gstack.
  ---
  dup gswap gdrop ;

: gover  ( 'gspace -- ) ( g: a b -- a b a )
  ---
    Leave a copy of the next most accessible gstack entry on top
    of the gstack.  The record is not copied.
  ---
  ?GSTACK-2CELLS-UNDER dup >gsp @ cell+ swap g@ ;

: gtuck  ( 'gspace -- ) ( g: a b -- b a b )
  ---
    Copy the top gstack item just below the second item.  The
    record is not copied.
  ---
  dup gswap gover ;

: g2drop  ( 'gspace -- ) ( g: a b --  )
  ---
    Drop the two topmost gstack entries, marking them as garbage
    and freeing them if appropriate.
  ---
  dup gdrop gdrop ;

: g3drop  ( 'gspace -- ) ( g: a b c --  )
  ---
    Drop the three topmost gstack entries, marking them as garbage
    and freeing them if appropriate.
  ---
  dup gdrop g2drop ;

: g2dup  ( 'gspace -- ) ( g: a b -- a b a b )
  ---
    Leave copies of the two topmost gstack entries.  The records
    are not copied.
  ---
  ?GSTACK-2CELLS-UNDER
  [ 2 cells ] literal over ?GROOM
  dup >r >gsp @ dup cell+ r@ g@ r> g@ ;

: gpick  ( u 'gspace -- ) ( g: a_u ... a_0 -- a_u ... a_0 a_u )
  ---
    Copy the u-th gstack entry to the top of the gstack. The
    record is not copied.  Throw an error if the input gstack
    does not have at least u+1 items.
  ---
  >r ( u) cells dup cell+ r@ ?GSTACK-UNDER
  r@ >gsp @ + r> g@ ;

: gexchange  ( i j 'gspace -- )
	     ( g: maxth ... minth ... -- minth ... maxth ... )
(
Exchange the ith and jth entries on the gstack, where the
top is the 0th.  Throw an error if there are not at least
max[i,j] + 1 entries.  Neither record is copied.
)
  >r cells swap cells
  2dup u> IF swap THEN ( min.cells max.cells)
  dup cell+ r@ ?GSTACK-UNDER
  2dup = IF 2drop rdrop EXIT THEN
  ( max.cells) r@ >gsp @ + ( &deep) dup @ rot
  ( &deep deep$ min.cells) r@ >gsp @ + ( &shal) dup @
  r>
  LOCALS| 'gspace shal &shal deep &deep |

  shal deep = IF EXIT THEN
  shal &deep !   deep &shal !
(
  Back links to gvars or to deeper than &deep do not change.
)
  shal 'gspace is-grec IF
    shal cell- @ ( blink) &shal &deep within
    IF &deep shal cell- ( blfa) ! THEN
  THEN

  deep 'gspace is-grec IF
    deep cell- @ ( blink) &deep =  \ was deepest copy
    IF
    (
      It was the deepest copy.  Scan for the deepest shallower
      copy, including that already stored at &shal.  Note that
      neg +LOOP includes limits.
    )
      &shal &deep cell-
      DO i @ deep = IF i LEAVE THEN -cell +LOOP
      ( &tween) deep cell- ( blfa) !
    THEN
  THEN ;

: fix-gblink  ( src targ 'gspace -- )
  ---
    Here src and targ are addresses of entries on the gstack.
    Assume that the 'grec formerly at src has been moved to
    targ, but that any binding has not changed yet.  If the grec
    is unbound, do nothing.  Else assume that if the grec was
    formerly bound to src, then there are no copies of 'grec
    deeper than targ.  If it was bound to src, bind it to targ
    instead.
  ---
  over @ swap is-grec IF
    ( targ) dup >r @ cell- ( src blfa) @ =
    IF r> ( targ) dup @ cell- ( blfa) ! EXIT THEN
    rdrop EXIT
  THEN 2drop ;

: grot  ( 'gspace -- ) ( g: a b c -- b c a )
  dup >r >gsp @ dup cell+ dup cell+ r@ gs3@ r>
  LOCALS| 'gspace c b a s2 s1 s0 |

  true CASE
    a b = OF
      a c <> IF			\ g: a a c   middle can't be deepest
	c s1 !   a s0 !		\ g: a c a
	s0 s1 'gspace fix-gblink 
      THEN
    ENDOF

    \ a <> b
    a c = OF			\ g: a b a   left and mid can be deepest
        a s1 !   b s2 !		\ g: b a a
        s1 s2 'gspace fix-gblink
        s2 s1 'gspace fix-gblink
    ENDOF

    \ a <> b, a <> c
    b c = OF			\ g: a b b
      a s0 !   b s2 !		\ g: b b a
      s2 s0 'gspace fix-gblink
      s1 s2 'gspace fix-gblink
    ENDOF

    \ all distinct		\ g: a b c
    a s0 !   b s2 !   c s1 !	\ g: b c a
    s2 s0 'gspace fix-gblink  
    s1 s2 'gspace fix-gblink
    s0 s1 'gspace fix-gblink
  ENDCASE
;

: -grot  ( 'gspace -- ) ( g: a b c -- c b a )
  dup >r >gsp @ dup cell+ dup cell+ r@ gs3@ r>
  LOCALS| 'gspace c b a s2 s1 s0 |

  true CASE
    a b = OF
      a c <> IF			\ g: a a c
	c s2 !   a s0 !		\ g: c a a
	s2 s1 'gspace fix-gblink
	s0 s2 'gspace fix-gblink
      THEN
    ENDOF

    \ a <> b
    a c = OF			\ g: a b a
        b s0 !   a s1 !		\ g: a a b
        s1 s0 'gspace fix-gblink
    ENDOF

    \ a <> b, a <> c
    b c = OF			\ g: a b b
      a s1 !   b s2 !		\ g: b a b
      s2 s1 'gspace fix-gblink
      s1 s2 'gspace fix-gblink
    ENDOF

    \ all distinct		\ g: a b c
    a s1 !   b s0 !   c s2 !	\ g: c a b
    s2 s1 'gspace fix-gblink  
    s1 s0 'gspace fix-gblink
    s0 s2 'gspace fix-gblink
  ENDCASE
;
